@@ -5,7 +5,7 @@ import pyvex
 from angr.engines.vex.claripy.irop import vexop_to_simop
 
 from .block import Block
-from .statement import Assignment, Store, Jump, ConditionalJump, Call, DirtyStatement
+from .statement import Assignment, Store, Jump, ConditionalJump, Call, DirtyStatement, Return
 from .expression import Const, Register, Tmp, DirtyExpression, UnaryOp, Convert, BinaryOp, Load, ITE
 
 l = logging.getLogger(name=__name__)
@@ -110,9 +110,15 @@ class VEXExprConverter(Converter):
             op1_val, op1_bits = operands[1].value, operands[1].bits
             operands[1] = Const(operands[1].idx, None, (1 << op1_bits) - op1_val, op1_bits)
 
+        signed = False
+        if op in {'CmpLE', 'CmpLT', 'CmpGE', 'CmpGT'}:
+            if vexop_to_simop(expr.op).is_signed:
+                signed = True
+
         return BinaryOp(manager.next_atom(),
                         op,
-                        operands
+                        operands,
+                        signed,
                         )
 
     @staticmethod
@@ -207,12 +213,52 @@ class VEXStmtConverter(Converter):
                                ins_addr=manager.ins_addr
                                )
 
+    @staticmethod
+    def LoadG(idx, stmt: pyvex.IRStmt.LoadG, manager):
+
+        sizes = {
+            'ILGop_Ident32': (32, 32, False),
+            'ILGop_Ident64': (64, 64, False),
+            'ILGop_IdentV128': (128, 128, False),
+            'ILGop_8Uto32': (8, 32, False),
+            'ILGop_8Sto32': (8, 32, True),
+            'ILGop_16Uto32': (16, 32, False),
+            'ILGop_16Sto32': (16, 32, True),
+        }
+
+        dst = VEXExprConverter.tmp(stmt.dst, manager.tyenv.sizeof(stmt.dst) // 8, manager)
+        load_bits, convert_bits, signed = sizes[stmt.cvt]
+        src = Load(manager.next_atom(),
+                   VEXExprConverter.convert(stmt.addr, manager),
+                   load_bits // 8,
+                   stmt.end,
+                   guard=VEXExprConverter.convert(stmt.guard, manager),
+                   alt=VEXExprConverter.convert(stmt.alt, manager))
+        if convert_bits != load_bits:
+            src = Convert(manager.next_atom(), load_bits, convert_bits, signed, src)
+
+        return Assignment(idx, dst, src, ins_addr=manager.ins_addr)
+
+    @staticmethod
+    def StoreG(idx, stmt: pyvex.IRStmt.StoreG, manager):
+
+        return Store(idx,
+                     VEXExprConverter.convert(stmt.addr, manager),
+                     VEXExprConverter.convert(stmt.data, manager),
+                     stmt.data.result_size(manager.tyenv) // 8,
+                     stmt.endness,
+                     guard=VEXExprConverter.convert(stmt.guard, manager),
+                     ins_addr=manager.ins_addr,
+                     )
+
 
 STATEMENT_MAPPINGS = {
     pyvex.IRStmt.Put: VEXStmtConverter.Put,
     pyvex.IRStmt.WrTmp: VEXStmtConverter.WrTmp,
     pyvex.IRStmt.Store: VEXStmtConverter.Store,
     pyvex.IRStmt.Exit: VEXStmtConverter.Exit,
+    pyvex.IRStmt.StoreG: VEXStmtConverter.StoreG,
+    pyvex.IRStmt.LoadG: VEXStmtConverter.LoadG,
 }
 
 
@@ -285,5 +331,13 @@ class IRSBConverter(Converter):
                                        ins_addr=manager.ins_addr
                                        )
                                   )
+        elif irsb.jumpkind == 'Ijk_Ret':
+            # return
+            statements.append(Return(manager.next_atom(),
+                                     VEXExprConverter.convert(irsb.next, manager),
+                                     [ ],
+                                     ins_addr=manager.ins_addr,
+                                     )
+                              )
 
         return Block(addr, irsb.size, statements=statements)
