@@ -6,6 +6,9 @@ class Expression(TaggedObject):
     """
     The base class of all AIL expressions.
     """
+
+    __slots__ = ('idx', )
+
     def __init__(self, idx, **kwargs):
         super(Expression, self).__init__(**kwargs)
         self.idx = idx
@@ -13,7 +16,13 @@ class Expression(TaggedObject):
     def __repr__(self):
         raise NotImplementedError()
 
-    def has_atom(self, atom):
+    def has_atom(self, atom, identity=True):
+        if identity:
+            return self is atom
+        else:
+            return self.likes(atom)
+
+    def likes(self, atom):
         return False
 
     def replace(self, old_expr, new_expr):
@@ -28,22 +37,29 @@ class Expression(TaggedObject):
         return r, replaced
 
     def __add__(self, other):
-        return BinaryOp(None, 'Add', [ self, other ])
+        return BinaryOp(None, 'Add', [ self, other ], False)
 
     def __sub__(self, other):
-        return BinaryOp(None, 'Sub', [ self, other ])
+        return BinaryOp(None, 'Sub', [ self, other ], False)
 
 
 class Atom(Expression):
-    def __init__(self, idx, variable, **kwargs):
+
+    __slots__ = ('variable', 'variable_offset', )
+
+    def __init__(self, idx, variable, variable_offset=0, **kwargs):
         super(Atom, self).__init__(idx, **kwargs)
         self.variable = variable
+        self.variable_offset = variable_offset
 
     def __repr__(self):
         return "Atom (%d)" % self.idx
 
 
 class Const(Atom):
+
+    __slots__ = ('value', 'bits', )
+
     def __init__(self, idx, variable, value, bits, **kwargs):
         super(Const, self).__init__(idx, variable, **kwargs)
 
@@ -74,12 +90,19 @@ class Const(Atom):
 
 
 class Tmp(Atom):
+
+    __slots__ = ('tmp_idx', 'bits', )
+
     def __init__(self, idx, variable, tmp_idx, bits, **kwargs):
         super(Tmp, self).__init__(idx, variable, **kwargs)
 
         self.tmp_idx = tmp_idx
         self.bits = bits
 
+    @property
+    def size(self):
+        return self.bits // 8
+
     def __repr__(self):
         return str(self)
 
@@ -96,6 +119,9 @@ class Tmp(Atom):
 
 
 class Register(Atom):
+
+    __slots__ = ('reg_offset', 'bits', )
+
     def __init__(self, idx, variable, reg_offset, bits, **kwargs):
         super(Register, self).__init__(idx, variable, **kwargs)
 
@@ -106,6 +132,11 @@ class Register(Atom):
     def size(self):
         return self.bits // 8
 
+    def likes(self, atom):
+        return type(self) is type(atom) and \
+                self.reg_offset == atom.reg_offset and \
+                self.bits == atom.bits
+
     def __repr__(self):
         return str(self)
 
@@ -118,26 +149,36 @@ class Register(Atom):
             return "%s" % str(self.variable.name)
 
     def __eq__(self, other):
-        return type(self) is type(other) and \
-               self.reg_offset == other.reg_offset and \
-            self.bits == other.bits
+        return self.likes(other) and self.idx == other.idx
 
     def __hash__(self):
-        return hash(('reg', self.reg_offset, self.bits))
+        return hash(('reg', self.reg_offset, self.bits, self.idx))
 
 
 class Op(Expression):
+
+    __slots__ = ('op', )
+
     def __init__(self, idx, op, **kwargs):
         super(Op, self).__init__(idx, **kwargs)
         self.op = op
 
+    @property
+    def verbose_op(self):
+        return self.op
+
 
 class UnaryOp(Op):
-    def __init__(self, idx, op, operand, **kwargs):
+
+    __slots__ = ('operand', 'bits', 'variable', 'variable_offset', )
+
+    def __init__(self, idx, op, operand, variable=None, variable_offset=None, **kwargs):
         super(UnaryOp, self).__init__(idx, op, **kwargs)
 
         self.operand = operand
         self.bits = operand.bits
+        self.variable = variable
+        self.variable_offset = variable_offset
 
     def __str__(self):
         return "(%s %s)" % (self.op, str(self.operand))
@@ -166,8 +207,15 @@ class UnaryOp(Op):
     def operands(self):
         return [ self.operand ]
 
+    @property
+    def size(self):
+        return self.bits // 8
+
 
 class Convert(UnaryOp):
+
+    __slots__ = ('from_bits', 'to_bits', 'is_signed', )
+
     def __init__(self, idx, from_bits, to_bits, is_signed, operand, **kwargs):
         super(Convert, self).__init__(idx, 'Convert', operand, **kwargs)
 
@@ -204,66 +252,119 @@ class Convert(UnaryOp):
 
 
 class BinaryOp(Op):
+
+    __slots__ = ('operands', 'bits', 'signed', 'variable', 'variable_offset', )
+
     OPSTR_MAP = {
         'Add': '+',
         'Sub': '-',
+        'Mul': '*',
+        'Div': '/',
+        'Xor': '^',
+        'And': '&',
+        'LogicalAnd': '&&',
+        'Or': '|',
+        'LogicalOr': '||',
+        'Shl': '<<',
+        'Shr': '>>',
+        'Sar': '>>a',
         'CmpEQ': '==',
         'CmpNE': '!=',
         'CmpLT': '<',
         'CmpLE': '<=',
         'CmpGT': '>',
         'CmpGE': '>=',
+        'CmpLTs': '<s',
+        'CmpLEs': '<=s',
+        'CmpGTs': '>s',
+        'CmpGEs': '>=s',
     }
 
-    def __init__(self, idx, op, operands, **kwargs):
+    def __init__(self, idx, op, operands, signed, variable=None, variable_offset=None, **kwargs):
         super(BinaryOp, self).__init__(idx, op, **kwargs)
 
         assert len(operands) == 2
         self.operands = operands
         self.bits = operands[0].bits if type(operands[0]) is not int else operands[1].bits
+        self.signed = signed
+        self.variable = variable
+        self.variable_offset = variable_offset
 
         # TODO: sanity check of operands' sizes for some ops
         # assert self.bits == operands[1].bits
 
     def __str__(self):
-        op_str = self.OPSTR_MAP.get(self.op, self.op)
+        op_str = self.OPSTR_MAP.get(self.verbose_op, self.verbose_op)
         return "(%s %s %s)" % (str(self.operands[0]), op_str, str(self.operands[1]))
 
     def __repr__(self):
-        return "%s(%s, %s)" % (self.op, self.operands[0], self.operands[1])
+        return "%s(%s, %s)" % (self.verbose_op, self.operands[0], self.operands[1])
 
     def __eq__(self, other):
         return type(other) is BinaryOp and \
                self.operands == other.operands and \
                self.op == other.op and \
-               self.bits == other.bits
+               self.bits == other.bits and \
+               self.signed == other.signed
 
     def __hash__(self):
-        return hash((self.op, tuple(self.operands), self.bits))
+        return hash((self.op, tuple(self.operands), self.bits, self.signed))
 
-    def has_atom(self, atom):
+    def has_atom(self, atom, identity=True):
         for op in self.operands:
-            if op == atom or op.has_atom(atom):
+            if identity and op == atom:
+                return True
+            if not identity and op.likes(atom):
+                return True
+            if op.has_atom(atom, identity=identity):
                 return True
         return False
 
     def replace(self, old_expr, new_expr):
-        r0, replaced_operand_0 = self.operands[0].replace(old_expr, new_expr)
-        r1, replaced_operand_1 = self.operands[1].replace(old_expr, new_expr)
+        if self.operands[0] == old_expr:
+            r0 = True
+            replaced_operand_0 = new_expr
+        else:
+            r0, replaced_operand_0 = self.operands[0].replace(old_expr, new_expr)
+
+        if self.operands[1] == old_expr:
+            r1 = True
+            replaced_operand_1 = new_expr
+        else:
+            r1, replaced_operand_1 = self.operands[1].replace(old_expr, new_expr)
 
         if r0 or r1:
-            return True, BinaryOp(self.idx, self.op, [ replaced_operand_0, replaced_operand_1 ], **self.tags)
+            return True, BinaryOp(self.idx, self.op, [ replaced_operand_0, replaced_operand_1 ], self.signed,
+                                  **self.tags)
         else:
             return False, self
 
+    @property
+    def verbose_op(self):
+        op = self.op
+        if self.signed:
+            op += "s"
+        return op
+
+    @property
+    def size(self):
+        return self.bits // 8
+
 
 class Load(Expression):
-    def __init__(self, idx, addr, size, endness, **kwargs):
+
+    __slots__ = ('addr', 'size', 'endness', 'variable', 'variable_offset', 'guard', 'alt', )
+
+    def __init__(self, idx, addr, size, endness, variable=None, variable_offset=None, guard=None, alt=None, **kwargs):
         super(Load, self).__init__(idx, **kwargs)
 
         self.addr = addr
         self.size = size
         self.endness = endness
+        self.guard = guard
+        self.alt = alt
+        self.variable = variable
+        self.variable_offset = variable_offset
 
     @property
     def bits(self):
@@ -275,10 +376,10 @@ class Load(Expression):
     def __str__(self):
         return "Load(addr=%s, size=%d, endness=%s)" % (self.addr, self.size, self.endness)
 
-    def has_atom(self, atom):
+    def has_atom(self, atom, identity=True):
         if type(self.addr) is int:
             return False
-        return self.addr.has_atom(atom)
+        return self.addr.has_atom(atom, identity=identity)
 
     def replace(self, old_expr, new_expr):
         r, replaced_addr = self.addr.replace(old_expr, new_expr)
@@ -292,13 +393,18 @@ class Load(Expression):
         return type(other) is Load and \
                self.addr == other.addr and \
                self.size == other.size and \
-               self.endness == other.endness
+               self.endness == other.endness and \
+               self.guard == other.guard and \
+               self.alt == other.alt
 
     def __hash__(self):
         return hash(('Load', self.addr, self.size, self.endness))
 
 
 class ITE(Expression):
+
+    __slots__ = ('cond', 'iffalse', 'iftrue', 'bits', )
+
     def __init__(self, idx, cond, iffalse, iftrue, **kwargs):
         super(ITE, self).__init__(idx, **kwargs)
 
@@ -313,8 +419,10 @@ class ITE(Expression):
     def __str__(self):
         return "((%s) ? (%s) : (%s))" % (self.cond, self.iftrue, self.iffalse)
 
-    def has_atom(self, atom):
-        return self.cond.has_atom(atom) or self.iftrue.has_atom(atom) or self.iffalse.has_atom(atom)
+    def has_atom(self, atom, identity=True):
+        return self.cond.has_atom(atom, identity=identity) or \
+               self.iftrue.has_atom(atom, identity=identity) or \
+               self.iffalse.has_atom(atom, identity=identity)
 
     def replace(self, old_expr, new_expr):
         cond_replaced, new_cond = self.cond.replace(old_expr, new_expr)
@@ -327,8 +435,15 @@ class ITE(Expression):
         else:
             return False, self
 
+    @property
+    def size(self):
+        return self.bits // 8
+
 
 class DirtyExpression(Expression):
+
+    __slots__ = ('dirty_expr', )
+
     def __init__(self, idx, dirty_expr, **kwargs):
         super(DirtyExpression, self).__init__(idx, **kwargs)
         self.dirty_expr = dirty_expr
@@ -349,11 +464,16 @@ class DirtyExpression(Expression):
 
 
 class BasePointerOffset(Expression):
-    def __init__(self, idx, bits, base, offset, **kwargs):
+
+    __slots__ = ('bits', 'base', 'offset', 'variable', 'variable_offset', )
+
+    def __init__(self, idx, bits, base, offset, variable=None, variable_offset=None, **kwargs):
         super().__init__(idx, **kwargs)
         self.bits = bits
         self.base = base
         self.offset = offset
+        self.variable = variable
+        self.variable_offset = variable_offset
 
     @property
     def size(self):
@@ -397,6 +517,9 @@ class BasePointerOffset(Expression):
 
 
 class StackBaseOffset(BasePointerOffset):
+
+    __slots__ = ()
+
     def __init__(self, idx, bits, offset, **kwargs):
         # stack base offset is always signed
         if offset >= (1 << (bits - 1)):
