@@ -1,10 +1,12 @@
 
 
-class TaggedObject(object):
+class TaggedObject:
     """
     A class that takes arbitrary tags.
     """
 
+    __slots__ = ('tags', )
+
     def __init__(self, **kwargs):
         self.tags = { }
         if kwargs:
